from pydantic import BaseModel, Field, validator
from typing import Union, Optional, List, Any, Dict
from enum import Enum
import datetime


# Data Source Definitions
class DataSource(BaseModel):
    format: str
    fieldSep: str
    data: str


# Attribute Specifications
class Attribute(BaseModel):
    name: str
    source: str
    type: str
    subType: str
    timeType: Optional[str] = None

# Operation Specification
class Operation(BaseModel):
    op: str
    inAttr: str
    outAttr: str
    windowUnit: str
    windowSize: Union[int, float]
    timeAttr: Optional[str] = None

# Configurations
class Config(BaseModel):
    numInitialStates: int = 12
    numHistogramBuckets: int = 10
    attributes: List[Attribute] = []
    ops: List[Operation] = []
    decTree_maxDepth: Optional[int] = 5
    decTree_minNormInfGainToSplit: Optional[float] = 0.1
    decTree_minEntropyToSplit: Optional[float] = 0.1
    ignoreConversionErrors: bool = False
    distWeightOutliers: Optional[float] = 0.05
    includeHistograms: bool = True
    includeDecisionTrees: bool = True
    includeStateHistory: bool = True

# Model Response and Related Structures
class Histogram(BaseModel):
    attrName: str
    freqSum: int
    freqs: List[int] = []
    bounds: Optional[List[float]] = None
    keys: Optional[List[Union[str, int]]] = None
    dayOfWeekFreqs: Optional[List[int]] = None
    monthFreqs: Optional[List[int]] = None
    hourFreqs: Optional[List[int]] = None

class DecisionTreeNode(BaseModel):
    nPos: int
    nNeg: int
    splitAttr: Optional[str] = None
    splitLabel: Optional[str] = None
    children: List['DecisionTreeNode'] = Field(default_factory=list)
    entropyBeforeSplit: float = 0.0
    splitCost: Optional[float] = None
    entropyAfterSplit: Optional[float] = None
    infGain: Optional[float] = None
    normInfGain: Optional[float] = None

class SuggestedLabel(BaseModel):
    label: str
    nCoveredInState: int
    nCoveredOutsideState: int
    nNotCoveredInState: int
    nNotCoveredOutsideState: int
    logOddsRatio: float

class State(BaseModel):
    stateNo: int
    initialStates: List[int]
    childStates: Optional[List[int]] = None
    parentState: Optional[int] = None
    sameAsParent: bool = False
    centroid: Optional[List[Dict]] = None
    stationaryProbability: Optional[float] = None
    nMembers: Optional[int] = None
    nextStateProbDistr: Optional[List[float]] = None
    histograms: Optional[List[Histogram]] = None
    xCenter: Optional[float] = None
    yCenter: Optional[float] = None
    radius: Optional[float] = None
    suggestedLabel: Optional[SuggestedLabel] = None
    decisionTree: Optional[DecisionTreeNode] = None

class Scale(BaseModel):
    nStates: int
    areTheseInitialStates: bool
    states: List[State]

class Column(BaseModel):
    name: str
    distWeight: float

class Dataset(BaseModel):
    cols: List[Column]

class Model(BaseModel):
    scales: List[Scale]
    totalHistograms: List[Histogram]
    stateHistoryTimes: List[float]
    stateHistoryInitialStates: List[int]
    config: Config
    dataset: Dataset

class ModelInfo(BaseModel):
    uuid: str
    username: str
    name: str
    description: Optional[str]
    dataset: str
    online: bool
    active: bool
    public: bool
    createdAt: datetime.datetime = Field(default_factory=datetime.datetime.now)
    model: Optional[Model] = None

    @validator('createdAt', pre=True, always=True)
    def convert_timestamp(cls, v):
        return datetime.datetime.fromtimestamp(v / 1000) if isinstance(v, (int, float)) else v

    # Additional validators can be added here.

class Response(BaseModel):
    status: str
    errors: List[str]
    model: Optional[Model] = None


class ClassifySamplesRequest(BaseModel):
    dataSource: DataSource
    model: Model


class BuildModelRequest(BaseModel):
    name: str
    description: str
    dataset: str
    public: bool
    dataSource: DataSource
    config: Config



import requests
import json
import pandas as pd
from streamstory.entities import Model, ModelInfo, BuildModelRequest, Config



class StreamStory:

    def __init__(self, api_url: str, api_key: str):
        self.api_url = api_url
        self.api_key = api_key


    def get_models(self) -> list[ModelInfo]:
        headers = {'x-api-key': self.api_key}
        response = requests.get(self.api_url + '/models', headers=headers)

        if response.status_code >= 300:
            raise Exception(f"Error: {response.status_code} - {response.text}")
        
        res_json =  response.json()['models']
        models = [ModelInfo(**model) for model in res_json]
        return models
  
  
    def get_model_by_uuid(self, uuid: str) -> ModelInfo:
        headers = {'x-api-key': self.api_key}
        response = requests.get(self.api_url + '/models/' + uuid, headers=headers)

        if response.status_code >= 300:
            raise Exception(f"Error: {response.status_code} - {response.text}")

        res_json =  response.json()
        model = ModelInfo(**res_json)
        return model

    def delete_model_by_uuid(self, uuid: str) -> None:
        headers = {'x-api-key': self.api_key}
        response = requests.delete(self.api_url + '/models/' + uuid, headers=headers)

        if response.status_code >= 300:
            raise Exception(f"Error: {response.status_code} - {response.text}")


    def build_model(self, build_model_req: BuildModelRequest) -> ModelInfo:
        headers = {'x-api-key': self.api_key}
        json_data = build_model_req.model_dump(mode='json')
        response = requests.post(self.api_url + '/models/build', headers=headers, json=json_data)
        
        if response.status_code >= 300:
            raise Exception(f"Error: {response.status_code} - {response.text}")
        
        return response.json() 


    def classify_datapoint(self, model_uuid, data):
        raise Exception("Not implemented yet")



import os
import time
import pandas as pd
from dotenv import load_dotenv
from streamstory.model import StreamStory
from streamstory.entities import BuildModelRequest, DataSource, Config, Attribute, Operation


if __name__ == "__main__":

    print("Running main.py")

    load_dotenv()

    api_url = os.getenv("STREAMSTORY_API_URL")
    api_key = os.getenv("STREAMSTORY_API_KEY")

    assert api_url is not None, "STREAMSTORY_API_URL not set"
    assert api_key is not None, "STREAMSTORY_API_KEY not set"

    print(f"api_url='{api_url}', api_key='{api_key}'")

    streamstory = StreamStory(api_url, api_key)

    data_path = "./data/river_flow.csv"  
    date_format = "%Y-%m-%dT%H:%M:%S"
    df = pd.read_csv(data_path, sep=";")

    df['datum'] = pd.to_datetime(df['datum'])
    df.set_index('datum', inplace=True)

    # resample by 1 day
    df = df.resample('1d').mean()
    # store the value of the previous day in a separate column and handle cases where no previous value 
    df['vrednost_prev_day'] = df['vrednost'].shift(1)	
    # drop rows with NaN values- Nan values are not allowed in the input data
    df = df.dropna()

    df['ss_timestamp'] = df.index.map(lambda x: time.mktime(x.timetuple()))

    data = ",".join(col for col in df.columns)

    for index, row in df.iterrows():
        data += "\n" + ",".join(str(row[col]) for col in df.columns)

    datasource = DataSource(
        format="csv",
        fieldSep=",",
        data=data, 
    )

    attributes = [
        Attribute(name="ss_timestamp", source="input", type="time", subType="integer", timeType="time"),
        Attribute(name="vrednost", source="input", type="numeric", subType="float"),
        Attribute(name="vrednost_prev_day", source="input", type="numeric", subType="float"),
    ]
   
    ops = [
        Operation(
            op="timeDelta",
            inAttr="vrednost",
            outAttr="river_flow_derivative",
            windowUnit="samples",
            windowSize=1
        ),
    ]
    
    config = Config(
        numInitialStates=12,
        numHistogramBuckets=10,
        attributes=attributes,
        ops=ops,
    )
    
    build_model_req = BuildModelRequest(
        name="River height",
        description="",
        dataset="river_height.csv",
        public=False,
        dataSource=datasource,
        config=config,
    )

    # build model
    model = streamstory.build_model(build_model_req)
    print(f"Succesfully built model")